use serde::{Serialize};

bitflags::bitflags! {
    /// InputTypeFlags are used to specify what type of input
    /// is to be allowed when in "assignment mode" for event mapping.
    /// This allows the caller to filter out unwanted event types in a
    /// generic way without having to handle the backend implementation details.
    #[derive(Default)]
    pub struct InputTypeFlags: u64
    {
        /// Any controller or mouse button including controller "hat" presses
        const Button      = 0b0000_0001;
        /// Any keyboard key
        const Key         = 0b0000_0010;
        /// Any controller axis
        const Axis        = 0b0000_0100;
        /// Any mouse movement
        const MouseMove   = 0b0000_1000;
        /// Key or Button combination
        const KeyOrButton = Self::Button.bits | Self::Key.bits;
        /// Axis, Key or Button combination
        const AxisKeyOrButton = Self::Axis.bits | Self::KeyOrButton.bits;
        /// Anything else (so you can do Any - Other for e.g. disconnect events etc.)
        const Other       = 0b0010_0000;
        /// Any type
        const Any = 0xffff_ffff_ffff_ffff;
    }
}

/// Describes states a key or button can be in
#[derive(Debug, PartialEq, Eq)]
pub enum ButtonState
{
    /// when key or button is pressed
    Pressed,
    /// when key or button is released
    Released,
}

/// Describes the source of input for a [ProcessingResult](struct.ProcessingResult.html)
#[derive(Debug, PartialEq)]
pub enum InputCause
{
    /// a key or button as source, with given state
    Button(ButtonState),
    /// an axis set to specific value in range <-1.0, 1.0>
    Axis(u8, f64), // axis, value TODO: map axis to actual possibilities
    /// a motion source (e.g. mouse) with x, y values
    Motion(f64, f64), // x, y
}

impl InputCause
{
    pub fn is_engaged(&self) -> bool {
        match self {
            InputCause::Button(state) => *state == ButtonState::Pressed,
            InputCause::Axis(_, value) => *value != 0.0,
            InputCause::Motion(_, _) => false, // never "engaged"
        }
    }

    pub fn is_axis(&self) -> bool {
        match self {
            InputCause::Axis(_, _) => true,
            _ => false,
        }
    }

    pub fn scalar_value(&self) -> f64 {
        match self {
            InputCause::Button(state) => if *state == ButtonState::Pressed {
                1.0
            } else {
                0.0
            },
            InputCause::Axis(_, value) => *value,
            InputCause::Motion(x, y) => x * y,
        }
    }
}

/// The result of a successful input processing by a Schema
#[derive(Debug, PartialEq)]
pub struct ProcessingResult<TUserAction>
where TUserAction: Clone + Serialize,
{
    /// The user defined action to be performed based on the input event
    pub action: TUserAction,
    /// The player_id if assigned to the Schema previously, otherwise None
    pub player_id: Option<usize>,
    /// The [InputCause](enum.InputCause.html) of this input, if known. None only if backend implementation is incomplete.
    pub cause: Option<InputCause>,
}

impl<TUserAction> ProcessingResult<TUserAction>
where TUserAction: Clone + Serialize,
{
    /// Specifies if the input cause is engaged (pressed or not equal to 0)
    /// *NOTE*: some input causes are never engaged, e.g. Motion
    pub fn input_is_engaged(&self) -> bool {
        if let Some(cause) = &self.cause {
            return cause.is_engaged()
        }

        false
    }

    /// Specifies if the input cause is an axis
    pub fn input_is_axis(&self) -> bool {
                if let Some(cause) = &self.cause {
            return cause.is_axis()
        }

        false
    }

    /// Reduces the input cause state value to a single f64
    /// Key or button states are either 0.0 or 1.0 for pressed
    /// Axis value stays same
    /// Motion x, y values are multiplied (x * y)
    pub fn input_scalar_value(&self) -> f64 {
        if let Some(cause) = &self.cause {
            return cause.scalar_value()
        }

        0.0
    }
}
